#!/bin/perl5
#!/apps/public/solaris/bin/perl -w

use strict;

#my $prefix = ".";
my $prefix = "/docs/www.all-day-breakfast.com/chinese";
my %cgivars = getcgivars();

# if $showchars, then we keep the characters in
my $showchars = $cgivars{"showchars"};

# if $showzy, then we write zhuyin (bopomofo) as well
my $showzy = $cgivars{"showzy"};

# and if $showpy, we write pinyin
my $showpy = $cgivars{"showpy"};

# if $showpy, we write english in the tooltips etc
my $showtrans = $cgivars{"showtrans"};

# if $do_links, link big5 characters to Web dictionary entries
my $do_links = $cgivars{"do_links"};

# if $do_audio, add .wav links
my $do_audio = $cgivars{"do_audio"};

# or some other format...
my $audioformat = $cgivars{"audioformat"};
defined($audioformat) or $audioformat = ".wav";

# where to put transl{iter}ations -- text, tooltips, status or atend
my $location = $cgivars{"location"};

# if $only_first, only the first occurrence of each word is translated
my $only_first = $cgivars{"only_first"};

# finally: if $text is defined, then we process that instead of our file
my $text = $cgivars{"text"};

# and $texttype tells us whether $text is "plain" or "html"
my $texttype = $cgivars{"texttype"};

### Now, we also have an alternative way of passing parameters -- the
### simplified form.  If we're using this, we get a single parameter
### from which we reconstruct the other parameters ourselves

my $simpleform = $cgivars{"simpleform"};
my $simpleaction = $cgivars{"simpleaction"};

if (defined($simpleform) and $simpleform) {
    $do_links = 1;
    $only_first = 1;
    if ($simpleaction eq "text") {
	$location = "text"; $showchars=1; $showpy = 1; $showzy = 0; $showtrans = 0;
    } elsif ($simpleaction eq "status") {
	$location = "status"; $showchars=1; $showpy = 1; $showzy = 0; $showtrans = 1;
    } elsif ($simpleaction eq "tooltips") {
	$location = "tooltips"; $showchars=1; $showpy = 1; $showzy = 1; $showtrans = 1;
    } elsif ($simpleaction eq "atend") {
	$location = "atend"; $showchars=1; $showpy = 1; $showzy = 0; $showtrans =1;
    } elsif ($simpleaction eq "nothing") {
	$location = "text"; $showchars = 1; $showpy = 0; $showzy = 0; $showtrans = 0; $do_links = 0; $only_first = 0;
    }
}

my $page;
if (defined($text)) { # we key off this later, to remove $page, so be careful
    $page = "$prefix/src/tmp$$";
    open DUMP, ">$page";
    if ($texttype eq "plain") {
	$text =~ s/&/&amp;/g; 
	$text =~ s/\</&lt\;/g;
	$text =~ s/\>/&gt\;/g;
	$text = "<PRE>$text</PRE>\n";
    }
    print DUMP $text;
    close DUMP;
} else {
    $page = $cgivars{"page"};
}

print "Content-type: text/html\n\n";

$showchars || $showzy || $showpy or
    $showchars = $showzy = $showpy = $showtrans = $do_links = $only_first = 1;
if (!defined($location)) { $location = "tooltips"; }


my $dict = "$prefix/big5dict";
my $index = "$prefix/big5dict.index";

open DICT, $dict;
open INDEX, $index;

my $minord;
seek INDEX, 0, 0;
read INDEX, $minord, length(pack "N");
$minord = unpack("N", $minord);

my $dictseeks = 0;
my $dicttests = 0;
my $dicthits = 0;
my $dictmisses = 0;
my $starttime = time;


my %big5array;
my %big5out;
my %englisharray;
my %englishout;
my %usedwords;

my %texthash; # if $location eq "atend"
my @textlist; # if $location eq "atend"

# this is only used if $do_tooltips
my %stylehash;
my $maxstyleargs = 1;

my %pages;
open(PAGELIST,"$prefix/pagelist") ||
    fail("Error: can't open page list \"$prefix/pagelist\"\n");
while (<PAGELIST>) {
    if (m/^([^ ]+) (.+)$/) {
	$pages{$1} = $2;
    }
}
close(PAGELIST);

my $pagename;
if (defined($pages{$page})) {
    $pagename = $pages{$page};
} else {
    if ($page) { $pagename = "Unknown page"; }
    else { $pagename = "Select a page"; }
}

print <<EOF;
<HTML><HEAD>
<META HTTP-EQUIV="content-type" CONTENT="text/html; charset=big5">
<TITLE>$pagename</TITLE>
EOF

if ($location eq "tooltips") {
    print <<EOF;
<SCRIPT LANGUAGE="JavaScript">
<!--
var NS4 = (document.layers) ? 1 : 0;
var IE4 = (document.all) ? 1 : 0;
function aEl() { }
function cEl() { }
// -->
</SCRIPT>
EOF
}

print <<EOF;
</HEAD>
<BODY BGCOLOR="#ffffff" TEXT="#000000" LINK="#8b0b04" VLINK="#052d6c" ALINK="#ab0b04"
EOF

print <<EOF;
>
<STYLE TYPE="text/css">
<!--
	A:link {text-decoration: none}
	A:visited {text-decoration: none}
	A:active {text-decoration: none}
-->
</STYLE>
EOF

if (!defined($text)) {

    print "<TABLE BGCOLOR=\"#d0d0f0\" WIDTH=\"100%\"><TR><TD BGCOLOR=\"#d0d0f0\" ALIGN=\"MIDDLE\">\n";
    print "<BR><FORM METHOD=\"GET\" ACTION=\"big5.cgi\">\n";
    
    print "<SELECT NAME=\"page\">\n";
    foreach (sort(keys(%pages))) {
	print "<OPTION VALUE=\"$_\"";
	print " SELECTED" if (defined($page) and $page eq $_);
	print ">$pages{$_}</OPTION>\n";
    }
    print "</SELECT>\n";
    print "<INPUT TYPE=\"submit\" VALUE=\"Display\"><!-- HR WIDTH=300 -->\n";

    if (defined($simpleform) and $simpleform) {

	print "<INPUT TYPE=\"hidden\" NAME=\"simpleform\" VALUE=1>\n";

	print "<BR><SELECT NAME=\"simpleaction\">\n";

	print "<OPTION VALUE=\"tooltips\"";
	print " SELECTED" if ($simpleaction eq "tooltips");
	print ">Add popup translations for Netscape 4 or IE4</OPTION>\n";

	print "<OPTION VALUE=\"status\"";
	print " SELECTED" if ($simpleaction eq "status");
	print ">Put translations in status-line with JavaScript</OPTION>\n";

	print "<OPTION VALUE=\"text\"";
	print " SELECTED" if ($simpleaction eq "text");
	print ">Add pinyin to the text</OPTION>\n";

	print "<OPTION VALUE=\"atend\"";
	print " SELECTED" if ($simpleaction eq "atend");
	print ">Append a word-list, to print for later study</OPTION>\n";

	print "<OPTION VALUE=\"nothing\"";
	print " SELECTED" if ($simpleaction eq "nothing");
	print ">Do nothing: present the text as plainly as possible</OPTION>\n";

	print "</SELECT>\n";

    } else {

	print "<TABLE><TR><TD VALIGN=MIDDLE>Show&nbsp;</TD><TD VALIGN=MIDDLE>\n";
	print "<TABLE BORDER=1 CELLPADDING=2 CELLSPACING=2><TR><TD>\n";
	
	print "<INPUT TYPE=\"checkbox\" NAME=\"showchars\" VALUE=\"1\"";
	if ($showchars) { print " CHECKED>\n"; } else { print ">\n"; }
	print "characters";
    
	print "<BR><INPUT TYPE=\"checkbox\" NAME=\"showpy\" VALUE=\"1\"";
	if ($showpy) { print " CHECKED>\n"; } else { print ">\n"; }
	print "pinyin, ";
    
	print "<INPUT TYPE=\"checkbox\" NAME=\"showzy\" VALUE=\"1\"";
	if ($showzy) { print " CHECKED>\n"; } else { print ">\n"; }
	print "zhuyin";
    
	print "<BR><INPUT TYPE=\"checkbox\" NAME=\"showtrans\" VALUE=\"1\"";
	if ($showtrans) { print " CHECKED>\n"; } else { print ">\n"; }
	print "translations";

	print "</TD></TR></TABLE>\n";
	print "</TD><TD VALIGN=MIDDLE>&nbsp;in&nbsp;</TD><TD VALIGN=MIDDLE>\n";
	print "<TABLE BORDER=1 CELLPADDING=2 CELLSPACING=2><TR><TD>\n";

	print "<INPUT TYPE=\"radio\" NAME=\"location\" VALUE=\"text\"";
	if ($location eq "text") { print " CHECKED>\n"; } else { print ">\n"; }
	print "text\n";

	print "<BR><INPUT TYPE=\"radio\" NAME=\"location\" VALUE=\"tooltips\"";
	if ($location eq "tooltips") { print " CHECKED>\n"; } else { print ">\n"; }
	print "tooltips\n";

	print "<BR><INPUT TYPE=\"radio\" NAME=\"location\" VALUE=\"status\"";
	if ($location eq "status") { print " CHECKED>\n"; } else { print ">\n"; }
	print "status line\n";

	print "<BR><INPUT TYPE=\"radio\" NAME=\"location\" VALUE=\"atend\"";
	if ($location eq "atend") { print " CHECKED>\n"; } else { print ">\n"; }
	print "separate wordlist\n";

	print "</TD></TR></TABLE>\n";
	print "</TD></TR></TABLE><TABLE><TR><TD>\n";
    
	print "<INPUT TYPE=\"checkbox\" NAME=\"do_links\" VALUE=\"1\"";
	if ($do_links) { print " CHECKED>\n"; } else { print ">\n"; }
	print "Link characters to online dictionaries";
    
	print "<BR><INPUT TYPE=\"checkbox\" NAME=\"only_first\" VALUE=\"1\"";
	if ($only_first) { print " CHECKED>\n"; } else { print ">\n"; }
	print "Only treat the first occurrence of each word";
    }    

    print "<BR><INPUT TYPE=\"checkbox\" NAME=\"do_audio\" VALUE=\"1\"";
    if ($do_audio) { print " CHECKED>\n"; } else { print ">\n"; }
    print "Include links to ";

    print "<SELECT NAME=\"audioformat\">\n";
    print "<OPTION VALUE=\".wav\"";
    if ($audioformat eq ".wav") { print " SELECTED"; }
    print ">WAV</OPTION>\n";
    print "<OPTION VALUE=\".aiff\"";
    if ($audioformat eq ".aiff") { print " SELECTED"; }
    print ">AIFF</OPTION>\n";
    print "<OPTION VALUE=\".au\"";
    if ($audioformat eq ".au") { print " SELECTED"; }
    print ">AU</OPTION>\n";
    print "</SELECT>\n";
    
    print " audio text-to-speech";

    if (!defined($simpleform) or !$simpleform) {
	print "<BR><SMALL>tooltips require NS4/IE4; status-line requires JavaScript</SMALL>\n";
    } else {
	print "<BR>&nbsp;\n";
    }

    print "</TD></TR></TABLE>\n";
    print "</FORM></TD></TR></TABLE><BR>\n";
}

if (defined($page) && $page) {
    open(PAGE, $page) || open(PAGE, $prefix . "/src/$page") ||
	fail("Error: can't open page \"$page\"\n");
} else {
    print "</BODY></HTML>\n";
    exit 0;
}

while (<PAGE>) {

    s,HREF="([^/]*)",HREF="big5.cgi?page=$1&showchars=$showchars&showzy=$showzy&showpy=$showpy&showtrans=$showtrans&do_links=$do_links&do_audio=$do_audio&audioformat=$audioformat&location=$location&only_first=$only_first&simpleform=$simpleform&simpleaction=$simpleaction",g;

#    if (!$do_english) { s,<DD>.*?</DD>,,; }

    my $pyphrase = "";
    my $big5phrase = "";
    my $tail = $_;

    while ($tail) {

	my $wl; # wordlength
	my ($bestpy, $bestchars, $bestenglish, $bl);
	my $chars = substr($tail, 0, 1);

	if ($chars lt "\200") {

	    if ($big5phrase) {
		if ($do_audio) {
		    $big5phrase = urlencode($big5phrase);
		    print " <A HREF=\"http://www.bell-labs.com/cgi-user/tts/mandarintts?text=$big5phrase&rate=slow&audio=$audioformat\"><IMG SRC=\"http://www.all-day-breakfast.com/chinese/audio.gif\" ALT=\"[$audioformat]\" BORDER=0 HSPACE=2></A>";
		}
		$big5phrase = "";
	    }

	    if ($pyphrase) {
		if ($location ne "text") {
		    $pyphrase = "";
		} else {
		    chop $pyphrase;
		    if ($showchars) {
			if ($showzy) {
			    print " ", py2zy($pyphrase);
			}
			if ($showpy) {
			    print "<small> [$pyphrase]</small>";
			}
		    } else {
			if ($showzy) {
			    print py2zy($pyphrase);
			    if ($showpy) {
				print " <small> [$pyphrase]</small>";
			    }
			} elsif ($showpy) {
			    print "<i>$pyphrase</i>";
			}
		    }
		    $pyphrase = "";
		    print " " if ($chars ne " ");
		}
	    }

	    print "$chars";
	    $tail = substr($tail, 1);
	    next;
	}

	if ($chars eq "\241") {

	    $bestchars = substr($tail, 0, 2);
	    $bestpy = $bestchars;
	    $bl = 2;

	} else {

	    for ($wl = 2; $wl <= 8; $wl += 2) {
		last if (substr($tail, $wl-2, $wl-1) lt "\200");
		$chars = substr($tail, 0, $wl);
		my $word = substr($tail, 0, $wl);

		if (!defined($big5array{$word})) {
		    my $dictword = lookup($word);
		    if ($dictword =~ m,^[^\s]+\s+\[(.*?)\]\s+/(.*?)/?\s*$,) {
			$big5array{$word} = $1;
			$englisharray{$word} = $2;
#			print STDERR "found $big5array{$word}, $englisharray{$word}\n";
		    } else {
#			print STDERR "refusing match of $word in $dictword\n";
		    }
		}

		if (defined($big5array{$word})) {
		    $bestchars = $word;
		    $bestpy = $big5array{$word};
		    $bestenglish = $englisharray{$word};
		    $bl = $wl;
		}
	    }
	}

	if (defined($bestpy)) {
	    $big5phrase .= $bestchars;
	    defined ($big5out{$bestchars}) or $big5out{$bestchars} = $bestpy;
	    defined ($englishout{$bestchars}) or $englishout{$bestchars} = $bestenglish;
	    if ($showchars || $location ne "text") {
		if ($usedwords{$bestchars}) {
		    print $bestchars;
		} else {
		    printchar($bestchars, $bestpy, $bestenglish);
		}
	    }
	    $pyphrase .= "$bestpy ";
	    $tail = substr($tail, $bl);
	    if ($only_first) { $usedwords{$bestchars} = 1; }
	} else {
	    $big5phrase .= substr($tail, 0, 2);
	    print substr($tail, 0, 2);
	    $tail = substr($tail, 2);
	}
    }
}

sub printchar {
    my $word = shift;
    my $py = shift;
    my $english = shift;
    my $js;

    my $stylename = "cE" . urlencode($word);
    $stylename =~ s/%//g;

    if ($location eq "atend") {
	if (!defined($texthash{$stylename}) and defined $english) {
	    my $ptext = "<TR><TD VALIGN=TOP>" . $word . "&nbsp;</TD>";
	    $ptext .= ($showpy ? "<TD VALIGN=TOP>$py</TD>" : "<TD>&nbsp;</TD>");
	    $ptext .= ($showzy ? ("<TD VALIGN=TOP>" . py2zy($py) . "</TD>") : "<TD>&nbsp;</TD>");
	    $ptext .= ($showtrans ? "<TD VALIGN=TOP>$english</TD>" : "<TD>&nbsp;</TD>");
	    $ptext .= "</TR>\n";
	    ($texthash{$stylename} = $ptext) =~ s,([^])\|,$1 / ,g;
	    push @textlist, $stylename;
	}
	print $word;
	return;
    }

    if ($location eq "tooltips") {
	if (!defined($stylehash{$stylename})) {
	    setstylehash($stylename, $word, $py, $english);
	}
    }

    if ($location eq "tooltips") {
	$js = "onMouseOver=\"aEl('$stylename', event)\" onMouseOut=\"cEl()\"";
    } elsif ($location eq "status") {
	$js = "onMouseOver=\"window.status='";
	if ($showpy) { $js .= $py; }
	if ($showtrans) { 
	    if ($showpy) {
		$js .= " - ";
	    }
	    $js .= "$english";
	}
	$js .= "'; return true;\" onMouseOut=\"window.status=''\"";
    }

    if ($do_links) {
	while ($word) {
	    my $char = substr($word, 0, 2);
	    $word = substr($word, 2);
	    if (substr($char, 0, 1) eq "\241") {
		print $char;
	    } else {
		my $uchar = urlencode($char);
		if ($location ne "text") {
		    print "<A HREF=\"http://zhongwen.com/cgi-bin/zi.cgi?b5=$uchar\" $js>$char</A>";
		} else {
		    print "<A HREF=\"http://zhongwen.com/cgi-bin/zi.cgi?b5=$uchar\">$char</A>";
		}
	    }
	}
    } else {
	if ($location ne "text" && substr($word, 0, 1) ne "\241") {
	    print "<A HREF=\"javascript:void(0)\" $js>$word</A>";
	} else {
	    print "$word";
	}
    }
}

sub setstylehash {
    my $stylename = shift;
    my $word = shift;
    my $py = shift;
    my $english = shift;

    # remove alternations from py (and hence zy) for multi-syllable
    # words -- otherwise it'll be too difficult to work out how to
    # write the zhuyin
    if ($word =~ m/^.../) { $py =~ s/\|.*$//; }

    my $count = length($word) / 2;
    my $pytail = $py;
    my $i;

    if ($count > $maxstyleargs) { $maxstyleargs = $count; }
    my $style = "makeEl$count(\"$stylename\", ";

    for ($i = 0; $i < $count; $i++) {
	my $zy;
	($zy = $pytail) =~ s/^([a-z:]+\d).*$/$1/;
	$zy = py2zy($zy);
	$zy =~ s/(..)/$1<BR>/g;
	$zy =~ s/<BR>$//;

	$style .= "\"" . substr($word, 0, 2) . "\", ";
	$style .= "\"$zy\", ";

	$pytail =~ s/^[a-z:]+\d(.*)$/$1/;
	$word = substr($word, 2);
    }

    $py =~ s,\|, / ,g;
    if (defined($english)) {
	$english =~ s,\|, / ,g;
    } else {
	$english = "";
    }
    $style .= "\"$py\", \"$english\");\n";
    $stylehash{$stylename} = $style;
    return;
}

if ($location eq "tooltips") {

    print <<EOF;
<SPAN ID="test" STYLE="position: absolute;"></SPAN>
<STYLE TYPE='text/css'>
<!--
	.tooltip {
		position: absolute;
		visibility: hidden;
		left: 0;
		top: 0;
		border-width: 2;
		border-color: #808020;
		border-style: outset;
		padding: 0;
		layer-background-color: beige;
		background-color: beige;
		width: 150;
	}
-->
</STYLE>
<SCRIPT LANGUAGE="JavaScript1.2">
<!--
    var style = ((NS4 && document.test) || IE4) ? 1 : 0;
    var timerID = null;
EOF

    foreach (keys(%stylehash)) {
	print "$stylehash{$_}";
    }

    my $i;
    for ($i = 1; $i <= $maxstyleargs; $i++) {
	my $j;
	print "function makeEl$i(id, ";
	for ($j = 0; $j < $i; ++$j) {
	    print "char$j, zy$j, ";
	}
	print "py, e) {\n";
	print "if (!style) return;\n";
	print "var str = \"<DIV CLASS=\'tooltip\' ID=\'\" + id + \"\'><TABLE><TR>\";\n";
	for ($j = 0; $j < $i; ++$j) {
	    if ($showchars) {
		print "str += \"<TD><FONT SIZE=6>\" + char$j + \"</FONT></TD>\";\n";
	    }
	    if ($showzy) {
		print "str += \"<TD><SMALL>\" + zy$j + \"</SMALL></TD>\";\n";
	    }
	}
	print "str += \"</TR></TABLE><TABLE CELLPADDING=2><TR><TD>\";\n";
	if ($showpy) {
	    if ($showtrans) {
		print "str += py + \"<BR>\" + e;\n";
	    } else {
		print "str += py;\n";
	    }
	} elsif ($showtrans) {
	    print "str += e;\n";
	}
	print "str += \"</TD></TR></TABLE></DIV>\";\n";
	print "document.write(str);\n}\n";
    }

    print <<EOF;
function displayEl(left, top) {
    if (NS4) document.releaseEvents(Event.MOUSEMOVE);
    document.onmousemove = null;
    var whichEl = (NS4) ? document[active] : document.all[active].style;
    if (!whichEl) return;
    if (NS4) {
	if (left > window.innerWidth + window.pageXOffset - 170) {
	    left = window.innerWidth + window.pageXOffset - 170;
	}
	if (top > window.innerHeight + window.pageYOffset - 130) {
	    top = window.innerHeight + window.pageYOffset - 130;
	}
    } else {
	if (left > document.body.clientWidth + document.body.scrollLeft - 170) {
	    left = document.body.clientWidth + document.body.scrollLeft - 170;
	}
	if (top > document.body.clientHeight + document.body.scrollTop - 130) {
	    top = document.body.clientHeight + document.body.scrollTop - 130;
	}
    }
    whichEl.left = left;
    whichEl.top = top;
    whichEl.visibility = (NS4) ? "show" : "visible";
}

function cEl() {
    if (!style) return;
    var whichEl = (NS4) ? document[active] : document.all[active].style;
    if (!whichEl) return;
    whichEl.visibility = (NS4) ? "hide" : "hidden";
    active = null;
    if (timerID) clearTimeout(timerID);
    if (NS4) document.releaseEvents(Event.MOUSEMOVE);
    document.onmousemove = null;
}

function aEl(id, e) {
    if (!style) return;
    active = id;
    if (NS4) document.captureEvents(Event.MOUSEMOVE);
    document.onmousemove = checkEl;
    checkEl(e);
}

function checkEl(e) {
    if (timerID) clearTimeout(timerID);
    var left = (NS4) ? e.pageX : event.clientX + document.body.scrollLeft;
    var top = (NS4) ? e.pageY + 20 : event.clientY + document.body.scrollTop + 20;
    timerID = setTimeout("displayEl(" + left + ", " + top + ")", 300);
}

// -->
</SCRIPT>
EOF
}

if ($location eq "atend") {
    print "<P><BR><CENTER><P><TABLE CELLPADDING=0 CELLSPACING=2 BORDER=0>\n";
    print "<TR><TD COLSPAN=4><B>Word list</B></TD></TR>\n";
    print "<TR><TD><IMG SRC=\"http://www.all-day-breakfast.com/dc.gif\" VSPACE=0 HSPACE=50 ALT=\"\"></TD><TD COLSPAN=3><IMG SRC=\"http://www.all-day-breakfast.com/dc.gif\" VSPACE=0 ALT=\"\"></TD></TR>";
    print map { $texthash{$_} } @textlist;
# or, for stroke order:
# print sort map { $texthash{$_} } keys(%texthash);
    print "</TABLE></CENTER>\n";
};

close PAGE;
if (defined($text)) {
    unlink($page); # temporary file
    print "\n<P><SMALL><I>";
    my $totaltime = time - $starttime;
    if ($totaltime == 1) { $totaltime .= " second"; }
    else { $totaltime .= " seconds"; }
    print "($totaltime, $dictseeks seeks, $dicttests comparisons, $dictmisses non-words tried, $dicthits distinct words found)</I></SMALL></P>\n";
}

print "</BODY></HTML>\n";

sub py2zy {
    local $_ = shift;
    if (m/^[\200-\377]/) { return $_; } # this is not py, don't convert
    s/([iu])e(\d)/$1$2/g;
    s/ui(\d)/uei$1/g;
    s/ing/ieng/g;
    s/ong/ueng/g;
    s/ang//g;
    s/eng//g;
    s/er//g;
    s/ai//g;
    s/ei//g;
    s/ao//g;
    s/ou//g;
    s/an//g;
    s/en//g;
    s/u:/uu/g;
    s/yu/uu/g;
    s/qu/quu/g;
    s/ju/juu/g;
    s/xu/xuu/g;
    s/yu/yuu/g;
    s/uu//g;
    s/ri/r/g;
    s/zi/z/g;
    s/ci/c/g;
    s/si/s/g;
    s/zhi/zh/g;
    s/chi/ch/g;
    s/shi/sh/g;
    s/yi/i/g;
    s/i//g;
    s/wu/u/g;
    s/u//g;
    s/w//g;
    s/a/\/g;
    s/o//g;
    s/e//g;
    s/y//g;
    s/l/\{/g;
    s/g/|/g;
    s/k/\}/g;
    s/j//g;
    s/q//g;
    s/x//g;
    s/zh//g;
    s/ch//g;
    s/sh//g;
    s/h/~/g;
    s/r//g;
    s/z//g;
    s/c//g;
    s/s//g;
    s/t/y/g;
    s/b/t/g;
    s/p/u/g;
    s/m/v/g;
    s/f/w/g;
    s/d/x/g;
    s/n/z/g;
    s/1//g;
    s/2//g;
    s/3//g;
    s/4//g;
    s/5/\/g;
    $_;
}


# function from "CGI Made Really Easy",
# http://www.jmarshall.com/easy/cgi/

sub getcgivars {
    my($in, %in) ;
    my($name, $value) ;


    # First, read entire string of CGI vars into $in
    if ( ($ENV{'REQUEST_METHOD'} eq 'GET') ||
         ($ENV{'REQUEST_METHOD'} eq 'HEAD') ) {
        $in= $ENV{'QUERY_STRING'} ;

    } elsif ($ENV{'REQUEST_METHOD'} eq 'POST') {
        if ($ENV{'CONTENT_TYPE'}=~ m#^application/x-www-form-urlencoded$#i) {
            length($ENV{'CONTENT_LENGTH'})
                || fail("No Content-Length sent with the POST request.") ;
            read(STDIN, $in, $ENV{'CONTENT_LENGTH'}) ;

        } else { 
            fail("Unsupported Content-Type: $ENV{'CONTENT_TYPE'}") ;
        }

    } else {
        fail("Script was called with unsupported REQUEST_METHOD.") ;
    }
    
    # Resolve and unencode name/value pairs into %in
    foreach (split('&', $in)) {
        s/\+/ /g ;
        ($name, $value)= split('=', $_, 2) ;
        $name=~ s/%(..)/chr(hex($1))/ge ;
        $value=~ s/%(..)/chr(hex($1))/ge ;
        $in{$name}.= "\0" if defined($in{$name}) ;  # concatenate multiple vars
        $in{$name}.= $value ;
    }

    return %in ;
}

sub urlencode
{
    my $arg = shift;
    $arg =~ s/([\200-\377+=&\\,\/.])/sprintf("%%%2x",ord($1))/ge;
    $arg =~ s/ /+/g;
    $arg;
}

sub fail
{
    my $message = shift;
    print "<HTML><HEAD><TITLE>Error</TITLE></HEAD><BODY>\n";
    print "<H1>Script error</H1><P>$message</P></BODY></HTML>\n";
    if (defined($text)) { unlink($page); }
    exit 0;
}

sub lookup
{
    my $answer = "failed";
    my $word = shift;
    my $idxchar = substr($word, 0, 2);
    my $idxlen = length($word) / 2;
    if ($idxlen > 5) { $idxlen = 5; }
    if ($idxlen < 1) { $dictmisses++; return $answer; } # undefined

#    print STDERR "Looking for $word (char $idxchar, length $idxlen)\n";
    
    my $longsize = length(pack "N");
    my $buffer;

    my $idxord =
	(ord(substr($idxchar, 0, 1)) - 128) * 256 +
	    ord(substr($idxchar, 1, 1));
    $idxord = $idxord * 5 + $idxlen - 1;

#    print STDERR "ord is $idxord\n";

    seek INDEX, ($idxord - $minord + 1) * $longsize, 0;
    read INDEX, $buffer, $longsize;
    my $idxseek = unpack("N", $buffer);
    $dictseeks++;

#    print STDERR "read seek position $idxseek from ",
#    ($idxord - $minord + 1) * $longsize, "\n";

    if ($idxseek == 0) { $dictmisses++; return $answer; }
    else { $idxseek--; }
    seek DICT, $idxseek, 0;
    $dictseeks++;

    my $terms = 0;
    while (<DICT>) {
	$terms++;
	$dicttests++;
	if (substr($_, 0, 2) ne $idxchar) {
	    $dictmisses++;
#	    print STDERR "failed after $terms terms\n";
	    return $answer;
	}
	if (substr($_, 0, $idxlen*2) eq $word) {
	    $dicthits++;
#	    print STDERR "succeeded after $terms terms, returning $_\n";
	    return $_; 
	}
    }

    $dictmisses++;
    return $answer;
}

